@@ -1,3 +1,4 @@
+#!/usr/bin/perl
 package OSGi::Osgish::ServerHandler;
 
 use strict;
@@ -5,6 +6,18 @@ use Term::ANSIColor qw(:constants);
 use OSGi::Osgish::Agent;
 use Data::Dumper;
 
+=head1 NAME 
+
+OSGi::Osgish::ServerHandler - Handler for coordinating the server access
+
+=head1 SYNOPSIS
+
+=head1 DESCRIPTION
+
+=head1 METHODS
+
+=cut
+
 sub new { 
     my $class = shift;
     my $osgish = shift || die "No osgish object given";
@@ -12,7 +25,7 @@ sub new {
                 osgish => $osgish
                };
     bless $self,(ref($class) || $class);
-    my $server = $self->_init_server_list($osgish->{config},$osgish->{args});
+    my $server = $self->_init_server_list($osgish->{initial_server},$osgish);
     $self->connect_to_server($server) if $server;
     return $self;
 }
@@ -68,18 +81,20 @@ sub list {
 
 sub _init_server_list {
     my $self = shift;
-    my $config = shift;
-    my $args = shift;
+    my $server = shift;
+    my $context = shift;
+    my $config = $context->{config};
+    my $args = $context->{args};
     my @servers = map { { name => $_->{name}, url => $_->{url}, from_config => 1 } } @{$config->get_servers};
     my $ret_server;
-    if ($args->{server}) {
-        my $config_s = $config->get_server_config($args->{server});
+    if ($server) {
+        my $config_s = $config->get_server_config($server);
         if ($config_s) {
             my $found = 0;
             my $i = 0;
-            my $entry = { name => $config_s->{name}, url => $config_s->{url}, from_config => 1 } ;
+            my $entry = { name => $server, url => $config_s->{url}, from_config => 1 } ;
             for my $s (@servers) {
-                if ($s->{name} eq $args->{server}) {
+                if ($s->{name} eq $server) {
                     $servers[$i] = $entry;
                     $found = 1;                 
                     last;
@@ -89,9 +104,9 @@ sub _init_server_list {
             push @servers,$entry unless $found;
             $ret_server = $config_s->{name};
         } else {
-            die "Invalid URL ",$args->{server} unless ($args->{server} =~ m|^\w+://|);
-            my $name = $self->_prepare_server_name($args->{server});
-            push @servers,{ name => $name, url => $args->{server} };
+            die "Invalid URL ",$server,"\n" unless ($server =~ m|^\w+://|);
+            my $name = $self->_prepare_server_name($server);
+            push @servers,{ name => $name, url => $server };
             $ret_server = $name;
         }
     }
@@ -158,4 +173,38 @@ sub _j4p_args {
     return $ret;
 }
 
+=head1 LICENSE
+
+This file is part of osgish.
+
+Osgish is free software: you can redistribute it and/or modify
+it under the terms of the GNU General Public License as published by
+the Free Software Foundation, either version 2 of the License, or
+(at your option) any later version.
+
+osgish is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+GNU General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with osgish.  If not, see <http://www.gnu.org/licenses/>.
+
+A commercial license is available as well. Please contact roland@cpan.org for
+further details.
+
+=head1 PROFESSIONAL SERVICES
+
+Just in case you need professional support for this module (or JMX or OSGi in
+general), you might want to have a look at www.consol.com Contact
+roland.huss@consol.de for further information (or use the contact form at
+http://www.consol.com/contact/)
+
+=head1 AUTHOR
+
+roland@cpan.org
+
+=cut
+
 1;
+
